#include "bayescc.h"
#include "common.h"

#include <boost/random.hpp>
#include <boost/random/normal_distribution.hpp>
#include <boost/math/distributions/uniform.hpp>


using namespace boost;

// interesting and confusing: http://code.google.com/p/ourmine/wiki/LectureNaiveBayes



BayesCC::BayesCC(Graph * _graph, double _alpha, double _beta, double _teta) :
        graph(_graph), alpha(_alpha), beta(_beta), teta(_teta)
{    

    // this may affect results....!!
    //alpha = ROUND2DECIMALS(alpha);
    //beta = ROUND2DECIMALS(beta);

    K = 1; // alpha as in http://en.wikipedia.org/wiki/Additive_smoothing
    M = 2; // m as in m-estimate: http://www.coli.uni-saarland.de/~crocker/courses/learning/lecture10.pdf

    edgeBeginProb = 0.1 * 1.0/3.0;
    edgeEndProb = 1.0/3.0;

    nLabels = graph->nLabels;

    changed = true;

    assert(alpha >= 0 && alpha <= 1);
    assert(beta >= 0 && beta <= 1);
    assert(teta >= 0 && teta <= 1);


}



BayesCC::~BayesCC() {
}




void BayesCC::initBayesNodes() {
    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);

    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];

        if(KNOWN_LABEL(node.labelId))
            assert(node.training);

        node.labelTime = 0;

        if(node.training) {
            node.bayes = true;
            node.P = 1;
        } else {
            node.bayes = false;
            node.P = 0; // never used. always overwritten.
        }

    }

    assert(inactiveEdges.empty());
    // xxx consider also other data structures, e.g. notFinalSet.


}

void BayesCC::labelRemainingNodesUsingPriors() {

    initLabelPriorsP();

    int count = 0;

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];

        if(KNOWN_LABEL(node.labelId))
            continue;

        //   cout << "WARNING: labelRemainingNodesUsingPriors on node id=" << node.id << "\n";

        node.labelId = topPrior();
        node.labelTime = iterationCount;

        count++;

    }

    cout << "Labeled " << count << " nodes using priors.\n";

}

set<VertexID> BayesCC::lookupInitialCandidates()  {
    set<VertexID> candidates;

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];

        if(KNOWN_LABEL(node.labelId)) {

            MyGraph::adjacency_iterator neighbourIt, neighbourEnd;
            boost::tie(neighbourIt, neighbourEnd) = adjacent_vertices(vertexID, graph->g);

            for (; neighbourIt != neighbourEnd; ++neighbourIt){
                VertexID neighborID = *neighbourIt;

                
                if(!graph->edge(vertexID,neighborID).active)
                    continue;

                Vertex & neighbor = graph->g[neighborID];

                if(UNKNOWN_LABEL(neighbor.labelId)) {
                    candidates.insert(*neighbourIt);
                    //          cout << "adding initial candidate node id=" << neighbor.id <<"\n";
                }

            }
        }
    }

    // the count of initial candidates does not change for any vale of beta < 1.    
    cout << "Count of initial candidate nodes: "<< candidates.size() << "\n";

    return candidates;
}





double BayesCC::naiveBayesP(int labelId, VertexID vertexID) {
    double P = log(labelPriorsP[labelId]);

    LOG("naiveBayesP labelId=" << labelId << " nodeId=" << graph->nodeVID(vertexID).id << "\n");

    //Vertex & node = graph->nodeVID(vertexID);

    MyGraph::adjacency_iterator neighbourIt, neighbourEnd;
    boost::tie(neighbourIt, neighbourEnd) = adjacent_vertices( vertexID, graph->g );

    for (; neighbourIt != neighbourEnd; ++neighbourIt){

        VertexID neighborID = *neighbourIt;
        Vertex & neighbor = graph->g[neighborID];


        Edge &edge = graph->edge(vertexID, neighborID);

        if(!neighbor.bayes)
            continue;

        LOG("Neighbor P=" << edge.P << " nodeId=" << neighbor.id << " labelId="  << neighbor.labelId << "\n");

        LOG("jointP=" << jointP[make_pair(neighbor.labelId, labelId)] << "\n");

        assert(KNOWN_LABEL(neighbor.labelId));
        assert(jointP[make_pair(neighbor.labelId, labelId)] != 0);

        // x^(1-y)  ranges between x and 1 depending on y. for y=0 => x, y=1 =>1.

        P +=  log(pow(jointP[make_pair(neighbor.labelId, labelId)],1-(edge.P*neighbor.P)));
        
        
    }

    P = pow(M_E,P);

    return P;
}





NodeLabelP BayesCC::labelNode(VertexID vertexID) {

    // label ids: 1 .. graph->labels.size()
    // labelId 0 means unknown label.

    Vertex & node = graph->g[vertexID];

    /*    if(node.labelId >0) {
       NodeLabelP nodeLabelP(vertexID,node.P);
      return nodeLabelP;
    }*/

    //cout << "Labeling node id= " << node.id <<" reachedTime= "  << node.reachedTime << "\n";

    assert(node.training == false); // we never ever label training nodes.

    int nLabels = graph->labels.size();
    double histogram[nLabels+1];
    double sum = 0;
    for(int i = 1; i <= nLabels; i++) {
        histogram[i] = naiveBayesP(i, vertexID);
        sum += histogram[i];
    }


    for(int i = 1; i <= nLabels; i++) {
        histogram[i] /= sum; // normalization..
        LOG("histogram[" << i << "]=" <<  histogram[i] << "\n");

    }


    int topLabel = 0;
    double topP = 0;

    int topLabel2 = 0;
    double topP2 = 0;

    for(int i = 1; i <= nLabels; i++) {
        if(histogram[i] > topP) {

            topLabel2 = topLabel;
            topP2 = topP;

            topP = histogram[i];
            topLabel = i;
        }
    }


    if(topLabel == 0) {
        topLabel = topPrior();
        topP = labelPriorsP[topLabel];
    }


    assert(topLabel != 0);

    changed |= node.labelId == topLabel;


    if(node.labelId != topLabel)
        node.labelTime = iterationCount;

    if(node.labelId != 0 && node.labelId != topLabel)
        nChanges++;

    if(node.labelId == 0)
        nLabeled++;


    node.labelId = topLabel;
    node.P = topP;

    if(alpha == 1) {
        node.bayes = 1; // use this label for all subsequent labelings. do not wait next iteration.
    }

    //if(topP == topP2) {
    //    cout << "Warning: node id=" << node.id << " labels same prob: topLabel=" <<  topLabel << " topLabel2=" << topLabel2 << " topP= " << topP << " topP2= " << topP2 << "\n";
    // }

    NodeLabelP nodeLabelP(vertexID,topP);
    return nodeLabelP;
}


int BayesCC::topPrior() {

    int topLabelId = 0;
    double topP = 0;

    for(unsigned int labelId = 1; labelId < labelPriorsP.size(); labelId++) {
        if(labelPriorsP[labelId] > topP) {
            topLabelId = labelId;
            topP = labelPriorsP[labelId];
        }
    }

    return topLabelId;

}

void BayesCC::updateJointProbabilities() {

    for(unsigned int i = 1; i <= graph->labels.size(); i++) {
        for(unsigned int j = 1; j <= graph->labels.size(); j++) {
            jointP[make_pair(i,j)] = 0;
        }
    }

    MyGraph::edge_iterator edgeIt, edgeEnd;
    boost::tie(edgeIt, edgeEnd) = edges(graph->g);

    double counts[graph->labels.size()+1];

    for(unsigned int i = 1; i <= graph->labels.size(); i++)
        counts[i] = 0;

    for (; edgeIt != edgeEnd; ++edgeIt){
        EdgeID edgeID = *edgeIt;
        Edge & edge = graph->g[edgeID];

        Vertex &node1 = graph->nodeFID(edge.node1);
        Vertex &node2 = graph->nodeFID(edge.node2);

        //       cout << "edge nodes: " << node1.id << " " << node2.id << " labels: " << node1.labelId << " " << node2.labelId <<  " active:" << node1.active << " " << node2.active << "\n";


        if(!edge.active)
            continue;

        if(!node1.bayes || !node2.bayes)
            continue;

        counts[node1.labelId]+= edge.P;
        counts[node2.labelId]+= edge.P;
        jointP[make_pair(node1.labelId,node2.labelId)]+= edge.P;
        jointP[make_pair(node2.labelId,node1.labelId)]+= edge.P;
    }


    for(unsigned int i = 1; i <= graph->labels.size(); i++) {
        for(unsigned int j = 1; j <= graph->labels.size(); j++) {

            double count = jointP[make_pair(i,j)];
            double prior = pcheck(labelPriorsP[i]);

            if(count > 0)  {
                jointP[make_pair(i,j)] = pcheck((count+(M*prior)) / (counts[j]+M));
            } else {
                jointP[make_pair(i,j)] = prior;
            }

            pcheck(jointP[make_pair(i,j)]);

        }
    }

}

inline double BayesCC::pcheck(double P) {

    if(P >1)  {
        assert(P <= 1+1e-5);
        P = 1;
    }

    if(P < 0) {
        assert(P >= -1e-5);
        P = 0;
    }

    return P;


}


void BayesCC::buildNodeRankPriorityList(set<VertexID> &candidates, NodeRankPriorityList &nodeRankPriorityList) {

    //    cout << "Building ranked list of candidates..\n";

    for(set<VertexID>::iterator iter = candidates.begin(); iter != candidates.end(); iter++) {

        //Vertex & node = graph->nodeVID(*iter);

        NodeRank nodeRank(*iter,0);

        MyGraph::adjacency_iterator neighbourIt, neighbourEnd;
        boost::tie(neighbourIt, neighbourEnd) = adjacent_vertices( *iter, graph->g );

        double count = 0;

        for (; neighbourIt != neighbourEnd; ++neighbourIt){
            VertexID neighborID = *neighbourIt;
            Vertex & neighbor = graph->g[neighborID];
            count++;

            if(neighbor.labelId != 0) {
                Edge & edge = graph->edge(nodeRank.vertexID, neighborID);
                nodeRank.r += edge.P * neighbor.P;
            }

        }

        nodeRank.r/= count;

        nodeRankPriorityList.push(nodeRank);

    }

    //cout << "Built, size= " << nodeRankPriorityList.size() << "\n";

}


bool BayesCC::iterate(set<VertexID> &candidates) {
    nChanges = 0;


    
    // update prior probabilities
    initLabelPriorsP();

    // update joint probabilies
    updateJointProbabilities();


    //cout << "nCandidates= " << candidates.size() << "\n";


    // update node labels

    for(set<VertexID>::iterator iter = notBayesSet.begin(); iter != notBayesSet.end(); iter++) {
        labelNode(*iter);
    }
    for(set<VertexID>::iterator iter = bayesSet.begin(); iter != bayesSet.end(); iter++) {
        labelNode(*iter);
    }


    NodeRankPriorityList  nodeRankPriorityList;
    buildNodeRankPriorityList(candidates, nodeRankPriorityList);

    if(teta < 1) {
        // assign labels to a fraction of the now-reachable nodes.

         int nToLabel = (int)MIN(((double)candidates.size()) * teta +1 , candidates.size());
         assert(nToLabel>=0);
        for( int i = 0; i < nToLabel; i++) {
            NodeRank nodeRank = nodeRankPriorityList.top();
            nodeRankPriorityList.pop();

            NodeLabelP nodeLabelP = labelNode(nodeRank.vertexID);
            notBayes.push(nodeLabelP);
            notBayesSet.insert(nodeLabelP.vertexID);
        }
        // cout << "Nodes labeled.\n";
    } else {


        // assign labels to now-reachable nodes.
        while(!nodeRankPriorityList.empty()) {

            //for(set<VertexID>::iterator iter = candidates.begin(); iter != candidates.end(); iter++) {

            NodeRank nodeRank = nodeRankPriorityList.top();
            nodeRankPriorityList.pop();

            NodeLabelP nodeLabelP = labelNode(nodeRank.vertexID);

            //cout << "Labeled new node id=" << graph->nodeVID(*iter).id << ": label id=" << graph->nodeVID(*iter).labelId << "\n";

            notBayes.push(nodeLabelP);
            notBayesSet.insert(nodeLabelP.vertexID);
        }

    }



    // add more labeled nodes to bayes learning model. +1 to addres limit conditions.
    int nBayes = (int)MIN(((double)notBayes.size()) * alpha +1 , notBayes.size());
    assert(nBayes >=0);

    double prevP = 1;// check priority list..

    for(int i = 0; i < nBayes; i++) {

        NodeLabelP nodeLabelP = notBayes.top();
        notBayes.pop();

        assert(prevP >= nodeLabelP.P); // check priority list..
        prevP = nodeLabelP.P;// check priority list..

        set<VertexID>::iterator iter = notBayesSet.find(nodeLabelP.vertexID);
        assert(iter != notBayesSet.end());
        notBayesSet.erase(iter);

        Vertex & node = graph->g[nodeLabelP.vertexID];
        node.bayes = true;
        bayesSet.insert(nodeLabelP.vertexID);
    }


    // move edges from inactive to active.  +1 to addres limit conditions.
    int nToMoveActive = (int)MIN((edgeActivableCount) * beta +1, edgeActivableCount);
    assert(nToMoveActive >= 0);

    edgeActivableCount-= nToMoveActive;
    assert(edgeActivableCount >= 0);


    double prevP2 = 1;// check priority list..

    for(int i = 0; i < nToMoveActive; i++) {
        EdgeP edgeP = inactiveEdges.top();
        inactiveEdges.pop();

        assert(prevP2 >= edgeP.P); // check priority list..
        prevP2 = edgeP.P;

        Edge & edge = graph->g[edgeP.edgeID];
        assert(edge.active == false);
        edge.active = true;
    }


    // build new list of reachable-nodes
    set<VertexID> newCandidates;
    lookupNextCandidates(candidates, newCandidates);

    candidates = newCandidates;

    // add candidates not labeled to the next iteration candidates.
    while(!nodeRankPriorityList.empty()) {
        NodeRank nodeRank = nodeRankPriorityList.top();
        nodeRankPriorityList.pop();
        candidates.insert(nodeRank.vertexID);
    }

    // termination criterion. continue?
    return (candidates.size() >0 || (notBayesSet.size() >0 && edgeActivableCount>0));

}

void BayesCC::lookupNextCandidates(set<VertexID> &candidates, set<VertexID> &newCandidates) {

    for(set<VertexID>::iterator iter = candidates.begin(); iter != candidates.end(); iter++) {

        VertexID vertexID = *iter; // dereference vertexIt, get the ID

        MyGraph::adjacency_iterator neighbourIt, neighbourEnd;
        boost::tie(neighbourIt, neighbourEnd) = adjacent_vertices( *iter, graph->g );

        for (; neighbourIt != neighbourEnd; ++neighbourIt){
            VertexID neighborID = *neighbourIt;

            if(!graph->edge(vertexID,neighborID).active)
                continue;

            Vertex & neighbor = graph->g[neighborID];
            if(UNKNOWN_LABEL(neighbor.labelId)) {
                newCandidates.insert(neighborID);

                //        cout << "adding new candidate node id=" << neighbor.id << "\n";
            }
        }

    }
}

void BayesCC::initLabelPriorsP() {

    labelPriorsP.resize(graph->labels.size()+1);

    for(unsigned int labelId = 1; labelId < labelPriorsP.size(); labelId++) {
        labelPriorsP[labelId] = 0;
    }

    double count = 0;

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];

        if(!node.bayes)
            continue;

        count++;
        labelPriorsP[node.labelId]+= node.P; //xxx era +1, uso .P similamente a quanto facciamo per calcolare le join probabilities.
    }

    for(unsigned int labelId = 1; labelId < labelPriorsP.size(); labelId++) {
        labelPriorsP[labelId] = (labelPriorsP[labelId]+K) / (count+K*nLabels);
        pcheck(labelPriorsP[labelId]);
    }


}


void BayesCC::initActiveEdges() {
    //double meanEdgeP = graph->meanEdgeP();


    MyGraph::edge_iterator edgeIt, edgeEnd;

    if(beta == 1) {

        cout << "initActiveEdges: beta= " << beta << "\n";

        boost::tie(edgeIt, edgeEnd) = edges(graph->g);
        for (; edgeIt != edgeEnd; ++edgeIt){
            EdgeID edgeID = *edgeIt;
            Edge & edge = graph->g[edgeID];
            edge.active = true;
        }

        edgeActivableCount = 0;

    } else {

        cout << "initActiveEdges: beta= " << beta << " activeEdgesRatio: " << edgeBeginProb << " -> " << edgeEndProb <<"\n";
        edgeActivableCount = ((double)num_edges(graph->g))*edgeEndProb;
        cout << "initActiveEdges: edgeActivableCount= "  << ((int)edgeActivableCount) << "\n";


        boost::tie(edgeIt, edgeEnd) = edges(graph->g);
        for (; edgeIt != edgeEnd; ++edgeIt){
            EdgeID edgeID = *edgeIt;
            Edge & edge = graph->g[edgeID];

            edge.active = false;
            EdgeP edgeP(edgeID, edge.P);
            inactiveEdges.push(edgeP);
        }


        // move edgeBeginProb% edges from inactive to active
        int nToMoveActive = (int)MIN(((double)inactiveEdges.size()) * edgeBeginProb +1, inactiveEdges.size());
        assert(nToMoveActive >=0);

        double prevP2 = 1;// check priority list..

        for(int i = 0; i < nToMoveActive; i++) {
            EdgeP edgeP = inactiveEdges.top();
            inactiveEdges.pop();

            assert(prevP2 >= edgeP.P); // check priority list..
            prevP2 = edgeP.P;

            Edge & edge = graph->g[edgeP.edgeID];
            assert(edge.active == false);
            edge.active = true;
        }

        edgeActivableCount = MAX(edgeActivableCount-nToMoveActive, 0);


    }


    cout << "initActiveEdges: after init, edgeActivableCount= "  << ((int)edgeActivableCount) << "\n";

    /*
    boost::tie(edgeIt, edgeEnd) = edges(graph->g);
    for (; edgeIt != edgeEnd; ++edgeIt){
        EdgeID edgeID = *edgeIt;
        Edge & edge = graph->g[edgeID];

        if(edge.active == false)
            cout << "yyy " << edge.P << "\n";
    }

    exit(0);
*/

}






void BayesCC::eval() {

    assert(0 && "non aggiornata");


    nLabeled = 0;
    iterationCount = 1;

    initBayesNodes();
    initActiveEdges();
    set<VertexID> candidates = lookupInitialCandidates();


    for(;;) {

        int c = candidates.size();
        bool ret = iterate(candidates);
        double ratioLabeled = nLabeled / ((double)num_vertices(graph->g));

        cout << "#ITER n= " << iterationCount << " c= " << c << " cNext= " << candidates.size() << " nLabeled= " << nLabeled << " nChanges= " << nChanges  << " ratioLabeled= " << ratioLabeled <<  " Acc= " << Evaluation::testAccuracy(graph) <<  " iAcc= " << Evaluation::testAccuracy(graph, iterationCount) << " inactivEdges= " << inactiveEdges.size() <<" notBayesNodes= " << notBayes.size() << "\n";

        iterationCount++;

        if(ret == false)
            break;
    }

    labelRemainingNodesUsingPriors();

}











