#include <boost/random/normal_distribution.hpp>
#include <boost/math/distributions/uniform.hpp>

#include "evaluation.h"
#include "random.h"

using namespace boost;

Evaluation::Evaluation(Graph *_graph) : graph(_graph)
{
}



void Evaluation::printConfusionMatrix() {
    vector< vector< double > > confusionMatrix;

    int nLabels = graph->labels.size();

    confusionMatrix.clear();
    confusionMatrix.resize(graph->labels.size()+1);
    for(unsigned int i = 0; i <= graph->labels.size(); i++) {
        confusionMatrix[i].resize(graph->labels.size()+1, 0);
    }

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
    double count = 0;
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];

        if(UNKNOWN_LABEL(node.gtLabelId))
            continue;

        if(node.training)
            continue;

        count++;

        assert(node.gtLabelId >= 1 && node.gtLabelId <= (int)graph->labels.size());
        assert(node.labelId >= 0 && node.labelId <= (int)graph->labels.size());

        confusionMatrix[node.gtLabelId][node.labelId]++;
    }

    cout << "\n\n";
    cout << "CONFUSION MATRIX BEGIN\n";

    cout << "&";
    for(int i = 1; i <= nLabels; i++) {
    cout << "$C_{" << i << "}$";
    if(i < nLabels)
        cout << "&";
    }

    cout << "\\\\ \\hline \\hline\n";

    for(int i = 1; i <= nLabels; i++) {
        cout << "$C_{" << i << "}$&";
        for(int j = 1; j <= nLabels; j++) {
            cout << "$" << (i == j ?  "\\mathbf{" : "") << confusionMatrix[i][j] << (i == j ? "}" : "") << "$" ;
            if(j < nLabels)
                cout << "&";
        }
        cout << "\\\\\n";
    }
    cout << "CONFUSION MATRIX END\n";
    cout << "\n\n";

    double N = 0;
    for(unsigned int i = 1; i <= graph->labels.size(); i++) {
        N += confusionMatrix[i][i];
    }

    double accuracy = N / count;

    cout << "#CONFUSION2 accuracy= " << accuracy << "\n";



}



double Evaluation::test() {

    vector< vector< double > > confusionMatrix;

    confusionMatrix.clear();
    confusionMatrix.resize(graph->labels.size()+1);
    for(unsigned int i = 0; i <= graph->labels.size(); i++) {
        confusionMatrix[i].resize(graph->labels.size()+1, 0);
    }

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
    double count = 0;
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];

        if(UNKNOWN_LABEL(node.gtLabelId))
            continue;


        if(node.training)
            continue;

        count++;

        //        if(!(node.labelId >= 0 && node.labelId <= graph->labels.size())) {
        //           cout << "ERRR " << node.labelId << " " << graph->labels.size() << "\n" ;
        //     }

        assert(node.gtLabelId >= 1 && node.gtLabelId <= (int)graph->labels.size());
        assert(node.labelId >= 0 && node.labelId <= (int)graph->labels.size());

        //    if(node.gtLabelId != node.labelId)
        //      cout << "mislabeled node id=" << node.id << " iteration= " << node.labelTime << " " << node.labelId << " " << node.gtLabelId <<" training=" << node.training << "\n";

        // very slow, can be improved a lot with a one-visit algorithm for all nodes.
        //       graph->cleanVisitedFlag();
        //     graph->reachesLabeledNode(vertexID);

        // if(node.labelTime == 0)
        //   assert(!graph->reachesLabeledNode(vertexID));

        confusionMatrix[node.gtLabelId][node.labelId]++;
    }

    double N = 0;
    for(unsigned int i = 1; i <= graph->labels.size(); i++) {
        N += confusionMatrix[i][i];
    }

    double accuracy = N / count;

    //    cout << "Accuracy=" << accuracy << " = " << N << "/" << count << "\n";

    return accuracy;

    // we compare nodes for which we know the label, that were not used for training.
}



void Evaluation::initTrainingSet() {

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];
        node.labelId = node.eLabelId;
        node.labelTime = -1;

    }

}

void Evaluation::setupTrainingSet(double P) {

    int trainingCount = 0;
    int testingCount = 0;
    int noLabelCount = 0;

    cout << "Initializing training/test sets... P=" << P << "\n";

    // we choose the nodes to be used for training.

    // labeled nodes are used for training with a probability of 0.5.

    rand48 rng = Random::getInstance();

    uniform_real<> ur(0.0,1.0);
    variate_generator<rand48&,uniform_real<> > drawSample(rng, ur);

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];

        if(KNOWN_LABEL(node.gtLabelId)) {

            //   if(node.id == 3 || node.id == 10 || node.id == 2 || node.id == 9) { /// XXX rimodificare commentando questa linea e scommentando quella sotto
            if(drawSample() <= P) { // if to be used for training...
                node.eLabelId = node.gtLabelId;
                node.training = true;
                trainingCount++;
            } else { // if to be used for testing...
                node.eLabelId = 0;
                node.training = false;
                testingCount++;
            }
        } else {
            node.eLabelId = 0;
            noLabelCount++;
            node.training = false;
        }
    }

    cout << "Using " << trainingCount << " nodes for training and " << testingCount << " for testing. There are " << noLabelCount << " unlabeled nodes.\n";

}



double Evaluation::testAccuracy(Graph *_graph) {

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(_graph->g);
    double N = 0;
    double count = 0;
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = _graph->g[vertexID];

        if(UNKNOWN_LABEL(node.gtLabelId))
            continue;

        if(node.training)
            continue;

        count++;

        assert(node.gtLabelId >= 1 && node.gtLabelId <= (int)_graph->labels.size());
        assert(node.labelId >= 0 && node.labelId <= (int)_graph->labels.size());

        if(node.gtLabelId == node.labelId)
            N++;
    }

    double accuracy = N / count;

    return accuracy;
}



double Evaluation::testAccuracy(Graph *_graph, int iterationCount) {

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(_graph->g);
    double N = 0;
    double count = 0;
    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = _graph->g[vertexID];

        if(node.labelTime != iterationCount)
            continue;

        if(UNKNOWN_LABEL(node.gtLabelId))
            continue;

        if(node.training)
            continue;

        count++;

        assert(node.gtLabelId >= 1 && node.gtLabelId <= (int)_graph->labels.size());
        assert(node.labelId >= 0 && node.labelId <= (int)_graph->labels.size());

        if(node.gtLabelId == node.labelId)
            N++;
    }

    if(count == 0)
        return -1;

    double accuracy = N / count;

    // normalize accuracy to its impact on the global problem:
    // accuracy *= ( count / ((double)num_vertices(_graph->g)) );
    // commented because not easy to understand in graphs.

    return accuracy;
}



void Evaluation::printAssignedLabels() {

    MyGraph::vertex_iterator vertexIt, vertexEnd;
    boost::tie(vertexIt, vertexEnd) = boost::vertices(graph->g);


    for (; vertexIt != vertexEnd; ++vertexIt){
        VertexID vertexID = *vertexIt; // dereference vertexIt, get the ID
        Vertex & node = graph->g[vertexID];

        if(KNOWN_LABEL(node.gtLabelId))
            continue;

        if(node.training)
            continue;

        cout << "Node nodeId= " << node.id << " gtLabelId= " << node.gtLabelId << " labelId= " << node.labelId << "\n";


    }

}
