#ifndef GRAPH_H
#define GRAPH_H

#include <map>
#include <vector>

#include <boost/tuple/tuple.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/visitors.hpp>
#include <boost/graph/breadth_first_search.hpp>
#include "histogram.h"

using namespace std;


#define UNKNOWN_LABEL(x) ((x) == 0)
#define KNOWN_LABEL(x) ((x) != 0)

//#define MAX_LABELS 100

struct Vertex{    

    int aid; // unique incremental node id, ranging [0,numberOfNodes-1]. to be used as array offset.
    int id; // unique node id (as in data files). not related to VertexID.
    string name; // node textual representation.
    int gtLabelId; // node label *ground truth*. 0 means unknown label.
    int eLabelId; // node label used during evaluation.
    int labelId; // node label assigned by algorithm. we compare this value to the ground truth! training samples have assignedLabelId=labelId.
    bool active; // true if node exists.
    bool training; // true if label used for training.

    // algorithm-specific.

    //vector<unsigned char> labelIds; // vector of labels determined by SamplingCC algorithm.
    bool bayes; // node used by BayesCC to evaluate class probability. Initially only training nodes.
    double P; // label probability
    bool visited;


    int labelTime;

};

struct Edge{
    double P; // weight probability
    int node1; // unique node id (as in data files). not related to VertexID.
    int node2; // unique node id (as in data files). not related to VertexID.
    bool active; // true if edge exists.    
};


// http://stackoverflow.com/questions/5056520/algorithm-for-selecting-all-edges-and-vertices-connected-to-one-vertex
// http://www.boost.org/doc/libs/1_50_0/libs/graph/doc/quick_tour.html
// http://www.boost.org/doc/libs/1_50_0/libs/graph/doc/graph_concepts.html
// http://www.boost.org/doc/libs/1_50_0/libs/graph/doc/adjacency_list.html
// http://www.boost.org/doc/libs/1_50_0/libs/graph/doc/using_adjacency_list.html


typedef boost::adjacency_list<  // adjacency_list is a template depending on :
        boost::listS,               //  The container used for egdes : here, std::list.
        boost::vecS,                //  The container used for vertices: here, std::vector.
        boost::undirectedS,           //  directed or undirected edges ?.
        Vertex,                     //  The type that describes a Vertex.
        Edge                        //  The type that describes an Edge
        > MyGraph;

/* From http://www.boost.org/doc/libs/1_47_0/libs/graph/doc/adjacency_list.html
In general, if you want your vertex and edge descriptors to be stable (never invalidated)
then use listS or setS for the VertexList and OutEdgeList template parameters of adjacency_list.
If you are not as concerned about descriptor and iterator stability, and are more concerned about
memory consumption and graph traversal speed, use vecS for the VertexList and/or OutEdgeList
template parameters. */


typedef MyGraph::vertex_descriptor VertexID;
typedef MyGraph::edge_descriptor   EdgeID;


class Graph
{
public:
    Graph(const char *pathname);
    Graph();
    
    map<int,string> labels;

    MyGraph g;
    string name;
    double nLabels;

    map<int,VertexID> nodeIds;

    void    loadFromFile(const char *pathname);
    void    dumpToNet(const char *pathname);
    void print();
    void printStats();
    void printComponents();
    void removeUnlabeledNodes();
    Edge &edge(VertexID id1, VertexID id2);
    Vertex &nodeVID(VertexID id);
    Vertex &nodeFID(int id);
    double meanEdgeP();
    unsigned long getFileSize(const char *pathname);
    bool reachesLabeledNode(VertexID vertexID);
    void cleanVisitedFlag();

    void setLabeledRatio(double ratio);

    void setRemoveRatioEdges(double ratio);


    Graph *sample(double ratioNodes);
    Vertex &insertNode(int id, string name, int labelId);
    Edge &insertEdge(int nodeId1, int nodeId2, double P);

    void copyLabels(Graph *graph);
    bool nodeIdExists(int nodeId);


private:
    void parseLabel(string str);
    void parseNode(string str);
    void parseEdge(string str);
    
};


#endif // GRAPH_H
